
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <sys/un.h>                         /* memmove() strcmp() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <linux/kd.h>                       /* KDSIGACCEPT */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdio.h>                          /* printf() */
#include <stdlib.h>                         /* free() exit() */
#include <sys/reboot.h>                     /* reboot() RB_DISABLE_CAD */
#include <assert.h>
#include <errno.h>

#include "initng_global.h"

#include "initng_active_db.h"
#include "initng_toolbox.h"
#include "initng_main.h"
#include "initng_execute.h"
#include "initng_common.h"
#include "initng_load_module.h"

#include "initng_handler.h"
#include "initng_kill_handler.h"

#include "initng_static_data_id.h"
#include "initng_static_states.h"


/*
 * This funtion is run from main when g.interrupt is set.
 * Will do special actions, for some status mode.
 */
void initng_run_interrupt_handlers(void)
{
    active_h *current, *q = NULL;

    S_;

    /* walk through all active services */
    while_active_db_safe(current, q)
    {
        assert(current->name);
        /* make sure it has a a_state set */
        if (!current->a_state)
            continue;

        /* call statehandler, now when we got an g.interrupt */
        if (current->a_state->state_handler)
            (*current->a_state->state_handler) (current);
        /* else try to set the next state */
        else if (current->a_state->next_state)
            mark_service(current, current->a_state->next_state);

        /* set time for last_interrput */
        current->last_interrupt = g.now;
    }
}

/* this function will send term signal to a service if it was marked 30 seconds or more */
void term_stopping_services(void)
{
    active_h *to_term, *q = NULL;
    process_h *process, *safe = NULL;
    int reactivate_alarm = FALSE;
    int timeout = SECONDS_BEFORE_KILL;

    S_;

    while_active_db_safe(to_term, q)
    {
        timeout = SECONDS_BEFORE_KILL;
        /* check if service is marked stopping */
        if (!IS_MARK(to_term, &STOPPING))
            continue;

        D_("Service %s is STOPPING\n", to_term->name);

        if (active_db_is(&STOP_TIMEOUT, to_term))
        {
            timeout = active_db_get_int(&STOP_TIMEOUT, to_term);
        }

        /* make sure this service has timed out */
        if ((time(0) - to_term->time_got_status) < timeout)
        {
            D_("Have to wait a bit more before killing this service!\n");
            reactivate_alarm = TRUE;
            continue;
        }

        D_("Service %s has been STOPPING for %i, and its more then %i seconds\n", to_term->name, (time(0) - to_term->time_got_status), timeout);

        /* ok, no go TERM all running processes */
        process = safe = NULL;
        while_processes_safe(process, safe, to_term)
        {
            if (process->pid < 0)
            {
                F_("There exists a process without a pid in the db!\n");
                continue;
            }


            /* check if the process still exits! */
            if (kill(process->pid, 0) < 0 && (errno == ESRCH))
            {
                /* This process is already dead */
                F_("process %s pid %i is already dead, cleaning up.\n",
                   to_term->name, process->pid);

                list_del(&process->list);
                process_db_free(process);

                mark_service(to_term, &STOPPED);
                /*handle_killed_by_pid(to_term->start_process->pid, 1); */
            }
            else
            {

                W_("Warning, terminating %s on pid %i!\n", to_term->name,
                   process->pid);
                kill(process->pid, SIGKILL);
            }
        }
    }

    /* reactivete the alarm if set */
    if (reactivate_alarm == TRUE)
        initng_set_sleep(timeout);
}

    /* M A R K    S T U F F */

    /* When this function is called, the service is marked for starting */
int start_service(active_h * service_to_start)
{
    assert(service_to_start);
    assert(service_to_start->name);

    S_;

    if (!service_to_start->a_state)
    {
        F_("Status for service not set.\n");
        return (FALSE);
    }

    if (!service_to_start->type)
    {
        F_("Type for service not set.\n");
        return (FALSE);
    }

    /* check system state, if we can launch. */
    if (g.sys_state != STATE_STARTING && g.sys_state != STATE_UP)
    {
        F_("Cant start service, when system status is: %i !\n", g.sys_state);
        return (FALSE);
    }

    /* it might alredy be starting */
    if (!IS_MARK(service_to_start, &LOADING)
        && (IS_STARTING(service_to_start) || IS_WAITING(service_to_start)))
    {
        D_("service %s is starting allredy.\n", service_to_start->name);
        return (TRUE);
    }

    /* it might alredy be up */
    if (IS_UP(service_to_start))
    {
        D_("service %s is alredy up!\n", service_to_start->name);
        return (TRUE);
    }

    /* make sure state is in the states, that we can start */
    if (!IS_MARK(service_to_start, &LOADING) && !IS_DOWN(service_to_start))
    {
        W_("cant set a service with status %i:%s, to start\n",
           service_to_start->a_state, service_to_start->a_state->state_name);
        return (FALSE);
    }

#ifdef NONOINTE
    /* it might not yet been stopped */
    if (IS_STOPPING(service_to_start))
    {
        /* remark it as started */
        D_("service %s did never have time to stop!\n",
           service_to_start->name);
        if (service_to_start->type == &TYPE_DAEMON)
            mark_service(service_to_start, &RUNNING);
        else
            mark_service(service_to_start, &DONE);
        return (TRUE);
    }
#endif



    if (service_to_start->type->start_service)
        return ((*service_to_start->type->start_service) (service_to_start));

    return (FALSE);
}

    /* S T O P     S E R V I C E */
int stop_service(active_h * service_to_stop)
{

    assert(service_to_stop);
    D_("stop_service(%s);\n", service_to_stop->name);

    if (!service_to_stop->a_state)
    {
        F_("A status is not set, so i cant stop_service\n");
        return (FALSE);
    }

    if (!service_to_stop->type)
    {
        F_("Service type is missing!\n");
        return (FALSE);
    }


    /* IF service is stopping, do nothing. */
    if (IS_STOPPING(service_to_stop))
        return (TRUE);

    /* if starting, direct abort */

    if (IS_STARTING(service_to_stop))
    {
        F_("Service is starting, cant stop it!\n");
        /*mark_service(service_to_stop, &STOPPED); */
        return (FALSE);
    }

    /* make sure the service is up */
    if (!IS_UP(service_to_stop))
    {
        /*W_("Cant stop service %s, with status %s\n",
           service_to_stop->name, service_to_stop->a_state->state_name);*/

	/* why we should mark it as fail stopping if it's not running at all?!? */
        mark_service(service_to_stop, &STOPPED);
        return (TRUE);
    }

    /* if stop_service code is included in type, use it. */
    if (service_to_stop->type->stop_service)
        return ((*service_to_stop->type->stop_service) (service_to_stop));

    return (FALSE);
}

int restart_service(active_h * service_to_restart)
{
    S_;
    if (!service_to_restart->a_state)
        return (FALSE);
    if (!service_to_restart->type)
        return (FALSE);

    /* if there exits a restart code, use it */
    if (service_to_restart->type->restart_service)
        return ((*service_to_restart->type->
                 restart_service) (service_to_restart));

    /* else, mark the service for restarting and stop it */
    active_db_set(&RESTARTING, service_to_restart);
    return (stop_service(service_to_restart));
}



/* Load new process named */
active_h *start_new_service_named(const char *service)
{
    active_h *to_load = NULL;

    D_("  --  %s\n", service);

    assert(service);

    if (strncmp(service, "virtual/", 8) == 0)
    {
        W_("Virtuals can not be started manually, they have to be provided.\n");
        return (NULL);
    }

    if ((to_load = active_db_find_by_name(service)))
    {
        D_("Service %s exits alredy!\n", to_load->name);
        return (to_load);
    }


    if (!(to_load = load_to_active(service)))
    {
        /* the function calling this function will print out an error */
        D_("Unable to load active for service %s\n", service);
        return (NULL);
    }

    /* IF service alredy got a state, its no good */
    if (to_load->a_state)
    {
        D_("Initial service alredy got a state!\n");
        return (NULL);
    }

    if (!mark_service(to_load, &LOADING))
    {
        D_("Mark service LOADING failed!\n");
        return (NULL);
    }

    /* okay, now start it */
    start_service(to_load);


    return (to_load);
}

/* enter a new runlevel */
int stop_all(void)
{
    active_h *ser, *q = NULL;

    S_;
    set_sys_state(STATE_STOPPING);

    while_active_db_safe(ser, q)
    {
        /* dont stop a stopping service */
        if (IS_STOPPING(ser))
            continue;

        /* stop services */
        stop_service(ser);
    }
    return (TRUE);
}
